#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv6.h>
#include <net/ipv6.h>
#include <net/ip.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/netlink.h>
#include <linux/ipv6.h>
#include <linux/icmp.h>
#include <linux/jhash.h>
#include <linux/spinlock.h>
#include <linux/netdevice.h>
#include <linux/sched.h>
#include <linux/sched/mm.h>

#define NETLINK_SOCKET     30
#define USER_PORT          100

#define NET_FLAG_ACCEPT     (unsigned int)0x00
#define NET_FLAG_LOG        (unsigned int)0x01
#define NET_FLAG_DROP       (unsigned int)0x02
#define NET_FLAG_REJECT     (unsigned int)0x04
#define NET_FLAG_MASK       (unsigned int)0x07

#define NET_WHITELIST       (unsigned int)0x10
#define NET_FLAG_WLOG       (unsigned int)(NET_FLAG_LOG | NET_WHITELIST)
#define NET_FLAG_WACCEPT    (unsigned int)(NET_FLAG_ACCEPT | NET_WHITELIST)
#define NET_FLAG_WDROP      (unsigned int)(NET_FLAG_DROP | NET_WHITELIST)
#define NET_FLAG_WREJECT    (unsigned int)(NET_FLAG_REJECT | NET_WHITELIST)

// net direction
// #define NET_DIR_INOUT       (unsigned int)0x000
#define NET_DIR_IN          (unsigned int)0x100     // only concerned with incoming packets
#define NET_DIR_OUT         (unsigned int)0x200     // only concerned with outgoing packets
#define NET_DIR_MASK        (unsigned int)0x300

#define NET_IN_LOG          (unsigned int)(NET_FLAG_LOG | NET_DIR_IN)
#define NET_IN_ACCEPT       (unsigned int)(NET_FLAG_ACCEPT | NET_DIR_IN)
#define NET_IN_DROP         (unsigned int)(NET_FLAG_DROP | NET_DIR_IN)

#define NET_OUT_LOG         (unsigned int)(NET_FLAG_LOG | NET_DIR_OUT)
#define NET_OUT_ACCEPT      (unsigned int)(NET_FLAG_ACCEPT | NET_DIR_OUT)
#define NET_OUT_DROP        (unsigned int)(NET_FLAG_DROP | NET_DIR_OUT)

// 带有ifname的 定义
#define NET_IFNAME          (unsigned int)0x400
#define NET_EXENAME         (unsigned int)0x800

static DEFINE_RWLOCK(my_rwlock);
struct sock *nl_sk = NULL;

struct ipv4_5tuple {
    u32 src_ip;
    u32 dst_ip;
    u16 src_port;
    u16 dst_port;
    u8 protocol;
};

struct ipv4_5tuple_ref {
    u8 reserve;
    u32 flags;
    u32 src_ip;
    u32 dst_ip;
    u32 src_ip_end;
    u32 dst_ip_end;
    u16 src_port;
    u16 dst_port;
    u16 src_port_end;
    u16 dst_port_end;
    u8 protocol;
    char name[TASK_COMM_LEN];  // process comm
};

struct ipv6_5tuple {
    struct in6_addr src_ip;
    struct in6_addr dst_ip;
    u16 src_port;
    u16 dst_port;
    u8 protocol;
};

struct ipv6_5tuple_ref {
    u8 reserve;
    u32 flags;
    struct in6_addr src_ip;
    struct in6_addr dst_ip;
    struct in6_addr src_ip_end;
    struct in6_addr dst_ip_end;
    u16 src_port;
    u16 dst_port;
    u16 src_port_end;
    u16 dst_port_end;
    u8 protocol;
    char name[TASK_COMM_LEN];  // process comm
};

struct nl_buf_ipv4 {
    u8  net_protocol;
    u32 data_len;
    int  pid;
    u64 timestamp;
    u64 start_time;
    struct ipv4_5tuple tuple;
};

struct nl_buf_ipv6 {
    u8  net_protocol;
    u32 data_len;
    int  pid;
    u64 timestamp;
    u64 start_time;
    struct ipv6_5tuple tuple;
};

/* Net monitor rules max len*/
#define MAX_RULES_LEN  200

struct netlink_rcv {
    size_t length; 
    union {
        struct {
            u8 net_family;
            u32 flags;
        };
        struct ipv4_5tuple_ref ipv4;
        struct ipv6_5tuple_ref ipv6;
    } rules[MAX_RULES_LEN];

};

struct filter {
    struct netlink_rcv  netc;
};

struct filter  *tuple_filter;
static int enable_filter = 0;

// 发送数据到用户层
static void send_to_user(char *msg, int len) {
    struct sk_buff *skb;
    struct nlmsghdr *nlh;

    skb = nlmsg_new(len, GFP_KERNEL);
    if (!skb) {
        printk(KERN_ERR "Failed to allocate skb\n");
        return;
    }

    nlh = nlmsg_put(skb, 0, 0, 0, len, 0);
    memcpy(nlmsg_data(nlh), msg, len);
    netlink_unicast(nl_sk, skb, USER_PORT, MSG_DONTWAIT);
}

void print_netc(struct netlink_rcv *netc)
{   
    int i;
    printk("Net context: length=%ld\n", netc->length);
    for(i = 0; i<netc->length; i++){

         if (netc->rules[i].net_family == 4) {

            printk("Rule %d: src %pI4--%pI4:%d--%d, dst %pI4--%pI4:%d--%d, protocol=%s fam %d, flags %x comm=%s\n", i+1,
                &netc->rules[i].ipv4.src_ip, &netc->rules[i].ipv4.src_ip_end,
                netc->rules[i].ipv4.src_port, netc->rules[i].ipv4.src_port_end,
                &netc->rules[i].ipv4.dst_ip, &netc->rules[i].ipv4.dst_ip_end,
                netc->rules[i].ipv4.dst_port, netc->rules[i].ipv4.dst_port_end,
                netc->rules[i].ipv4.protocol == 6 ? "TCP" : "UDP", netc->rules[i].net_family, netc->rules[i].flags, netc->rules[i].ipv4.name);
        }else{
            printk("Rule %d: src %pI6--%pI6:%d--%d, dst %pI6--%pI6:%d--%d, protocol=%s fam %d flags %x comm=%s\n", i+1,
                &netc->rules[i].ipv6.src_ip, &netc->rules[i].ipv6.src_ip_end,
                netc->rules[i].ipv6.src_port, netc->rules[i].ipv6.src_port_end,
                &netc->rules[i].ipv6.dst_ip, &netc->rules[i].ipv6.dst_ip_end,
                netc->rules[i].ipv6.dst_port, netc->rules[i].ipv6.dst_port_end,
                netc->rules[i].ipv6.protocol == 6 ? "TCP" : "UDP", netc->rules[i].net_family, netc->rules[i].flags, netc->rules[i].ipv6.name);
        }
    }
}

static int ipv4_addr_le(u32 addr1, u32 addr2)
{
    u32 addr1_n = htonl(addr1);
    u32 addr2_n = htonl(addr2);

    return addr1_n <= addr2_n;
}

static int ipv4_match(struct ipv4_5tuple t1, struct ipv4_5tuple_ref t2)
{
    int ret = 0;

    if(t1.protocol != t2.protocol && t1.protocol != IPPROTO_ICMP)
        return 0;

    ret = (t2.src_port == 0 || (t2.src_port <= t1.src_port && t1.src_port <= t2.src_port_end)) &&
          (t2.dst_port == 0 || (t2.dst_port <= t1.dst_port && t1.dst_port <= t2.dst_port_end)) &&
          (t2.src_ip == 0 || (ipv4_addr_le(t2.src_ip, t1.src_ip) && ipv4_addr_le(t1.src_ip, t2.src_ip_end))) &&
          (t2.dst_ip == 0 || (ipv4_addr_le(t2.dst_ip, t1.dst_ip) && ipv4_addr_le(t1.dst_ip, t2.dst_ip_end)));

    return ret;
}

// Ipv6 address comparison addr1 < addr2 return 1, otherwise return 0
static int ipv6_addr_le(struct in6_addr *addr1, struct in6_addr *addr2)
{
    int i;
    for (i = 0; i < 8; ++i) {
        uint16_t part1 = ntohs(addr1->s6_addr16[i]);
        uint16_t part2 = ntohs(addr2->s6_addr16[i]);

        if (part1 != part2) {
            return part1 < part2 ? 1 : 0;
        }
    }
    return 1;
}

static int ipv6_match(struct ipv6_5tuple t1, struct ipv6_5tuple_ref t2)
{
    int ret = 0;

    if(t1.protocol != t2.protocol && t1.protocol != IPPROTO_ICMPV6)
        return 0;

    ret = (t2.src_port == 0 || (t2.src_port <= t1.src_port && t1.src_port <= t2.src_port_end)) &&
            (t2.dst_port == 0 || (t2.dst_port <= t1.dst_port && t1.dst_port <= t2.dst_port_end)) &&
            (ipv6_addr_any(&t2.src_ip) || ipv6_addr_le(&t2.src_ip, &t1.src_ip) || ipv6_addr_le(&t1.src_ip, &t2.src_ip_end)) &&
            (ipv6_addr_any(&t2.dst_ip) || ipv6_addr_le(&t2.dst_ip, &t1.dst_ip) || ipv6_addr_le(&t1.dst_ip, &t2.dst_ip_end));


    return ret;
}

static int match_rulev4(struct ipv4_5tuple tuple, struct netlink_rcv *netc,
                        char *exe_name, unsigned int direction) 
{   
    int i;
    for (i = 0; i < netc->length; i++) {
        if(netc->rules[i].flags & NET_WHITELIST) {
            return NET_FLAG_ACCEPT;
        }
        if(netc->rules[i].net_family == 4) {
            if(ipv4_match(tuple, netc->rules[i].ipv4)) {
                int ret = netc->rules[i].flags & NET_FLAG_MASK;

                if(ret == netc->rules[i].flags)
                    return ret;

                if (netc->rules[i].flags & NET_EXENAME) {
                    if(strncmp(exe_name, netc->rules[i].ipv4.name, TASK_COMM_LEN) == 0) {
                        return ret;
                    }
                }
                if (netc->rules[i].flags & NET_DIR_IN) {
                    if (direction == NF_INET_PRE_ROUTING) {
                        return ret;
                    }
                } else if (netc->rules[i].flags & NET_DIR_OUT) {
                    if (direction == NF_INET_POST_ROUTING) {
                        return ret;
                    }
                }
            }
        }
    }
    return NET_FLAG_ACCEPT;
}

static int match_rulev6(struct ipv6_5tuple tuple, struct netlink_rcv *netc,
                        char *exe_name, unsigned int direction) 
{
    int i;
    for (i = 0; i < netc->length; i++) {
        if(netc->rules[i].flags & NET_WHITELIST) {
            return NET_FLAG_ACCEPT;
        }
        if(netc->rules[i].net_family == 6){
            if(ipv6_match(tuple, netc->rules[i].ipv6)){
                int ret = netc->rules[i].flags & NET_FLAG_MASK;

                if(ret == netc->rules[i].flags)
                    return ret;

                if (netc->rules[i].flags & NET_EXENAME) {
                    if(strncmp(exe_name, netc->rules[i].ipv4.name, TASK_COMM_LEN) == 0) {
                        return ret;
                    }
                }
                if (netc->rules[i].flags & NET_DIR_IN) {
                    if (direction == NF_INET_PRE_ROUTING) {
                        return ret;
                    }
                } else if (netc->rules[i].flags & NET_DIR_OUT) {
                    if (direction == NF_INET_POST_ROUTING) {
                        return ret;
                    }
                }
            }
        }
    }
    return NET_FLAG_ACCEPT;
}

static int ipv4_msg_event(struct task_struct *task, struct sk_buff *skb,
                            struct nl_buf_ipv4 *buf)
{
    struct iphdr *iph;
    struct tcphdr *tcph;
    struct udphdr *udph;
    unsigned int data_len;

    iph = ip_hdr(skb);
    data_len = skb->len - (iph->ihl * 4);
    buf->data_len = data_len;

    if (iph->protocol == IPPROTO_TCP) {
        tcph = tcp_hdr(skb);
        if(iph->saddr==iph->daddr)
            return -1;
        buf->tuple.src_ip = iph->saddr;
        buf->tuple.dst_ip = iph->daddr;
        buf->tuple.src_port = tcph->source;
        buf->tuple.dst_port = tcph->dest;
        buf->tuple.protocol = iph->protocol;
    }
    else if (iph->protocol == IPPROTO_UDP) {
        udph = udp_hdr(skb);
        buf->tuple.src_ip = iph->saddr;
        buf->tuple.dst_ip = iph->daddr;
        buf->tuple.src_port = udph->source;
        buf->tuple.dst_port = udph->dest;
        buf->tuple.protocol = iph->protocol;
    } else if (iph->protocol == IPPROTO_ICMP){
        struct icmphdr *icmph;
        icmph = icmp_hdr(skb);
        buf->tuple.src_ip = iph->saddr;
        buf->tuple.dst_ip = iph->daddr;
        buf->tuple.src_port = icmph->un.echo.id;
        buf->tuple.dst_port = icmph->un.echo.id;
        buf->tuple.protocol = iph->protocol;

    }
    else { 
        printk(KERN_INFO "Unknown protocol: %d\n", iph->protocol);
        return -1;
    }
    return 0;
}

static int ipv6_msg_event(struct task_struct *task, struct sk_buff *skb,
                            struct nl_buf_ipv6 *buf)

{
    struct ipv6hdr *ipv6h;
    struct tcphdr *tcph;
    struct udphdr *udph;
    unsigned int data_len;

    ipv6h = ipv6_hdr(skb);
    data_len = skb->len - sizeof(struct ipv6hdr);
    buf->data_len = data_len;
    if (ipv6h->nexthdr == IPPROTO_TCP) {
        tcph = tcp_hdr(skb);
        buf->tuple.src_ip = ipv6h->saddr;
        buf->tuple.dst_ip = ipv6h->daddr;
        buf->tuple.src_port = tcph->source;
        buf->tuple.dst_port = tcph->dest;
        buf->tuple.protocol = ipv6h->nexthdr;

    }
    else if (ipv6h->nexthdr == IPPROTO_UDP) {
        udph = udp_hdr(skb);
        buf->tuple.src_ip = ipv6h->saddr;
        buf->tuple.dst_ip = ipv6h->daddr;
        buf->tuple.src_port = udph->source;
        buf->tuple.dst_port = udph->dest;
        buf->tuple.protocol = ipv6h->nexthdr;

    } else if (ipv6h->nexthdr == IPPROTO_ICMPV6) {
        struct icmp6hdr *icmp6h;
        icmp6h = icmp6_hdr(skb);
        buf->tuple.src_ip = ipv6h->saddr;
        buf->tuple.dst_ip = ipv6h->daddr;
        buf->tuple.src_port = icmp6h->icmp6_identifier;
        buf->tuple.dst_port = icmp6h->icmp6_identifier;
        buf->tuple.protocol = ipv6h->nexthdr;

    } 
    else {
        printk(KERN_INFO "Unknown protocol: %d\n", ipv6h->nexthdr);
        return -1;
    }
    return 0;
}

// Netfilter钩子函数
static unsigned int nf_hook_func(void *priv, struct sk_buff *skb,
                                 const struct nf_hook_state *state) 
{
    struct task_struct *task;
    int ret;
    int match;
    char *exe_name = NULL;
    task = get_current();
    exe_name = task->comm;

    if(!enable_filter)
        return NF_ACCEPT;
    
    if(skb->protocol == htons(ETH_P_IP)) {
        struct nl_buf_ipv4 buf;
        buf.net_protocol = 4; //ipv4
        buf.pid = task->pid;
        buf.timestamp = jiffies_to_msecs(jiffies);
        buf.start_time =  task->start_time;

        ret = ipv4_msg_event(task, skb, &buf);

        read_lock_bh(&my_rwlock);
        match = match_rulev4(buf.tuple, &tuple_filter->netc, exe_name, state->hook);

        read_unlock_bh(&my_rwlock);

        switch(match) {
            case NET_FLAG_ACCEPT:
                return NF_ACCEPT;
            case NET_FLAG_LOG:
                send_to_user((char *)&buf, sizeof(buf));
                break;
            case NET_FLAG_DROP:
                return NF_DROP;
            default:
                break;
        }

    }
    else if(skb->protocol == htons(ETH_P_IPV6)) {
        struct nl_buf_ipv6 buf;
        buf.net_protocol = 6; //ipv6
        buf.pid = task->pid;
        buf.timestamp = jiffies_to_msecs(jiffies);
        buf.start_time =  task->start_time;
        ret = ipv6_msg_event(task, skb, &buf);

        read_lock_bh(&my_rwlock);
        match = match_rulev6(buf.tuple, &tuple_filter->netc, exe_name, state->hook);
        read_unlock_bh(&my_rwlock);

        switch(match) {
            case NET_FLAG_ACCEPT:
                return NF_ACCEPT;
            case NET_FLAG_LOG:
                send_to_user((char *)&buf, sizeof(buf));
                break;
            case NET_FLAG_DROP:
                return NF_DROP;
            default:
                break;
        }
    }
    return NF_ACCEPT;
}

// 注册Netfilter钩子
static struct nf_hook_ops nfho_ipv4_pre = {
    .hook = nf_hook_func,
    .hooknum = NF_INET_PRE_ROUTING,
    .pf = PF_INET,
    .priority = NF_IP_PRI_FIRST,
};

static struct nf_hook_ops nfho_ipv4_post = {
    .hook = nf_hook_func,
    .hooknum = NF_INET_POST_ROUTING,
    .pf = PF_INET,
    .priority = NF_IP_PRI_FIRST,
};

// 注册Netfilter钩子
static struct nf_hook_ops nfho_ipv6_pre = { // added for IPv6
    .hook = nf_hook_func,
    .hooknum = NF_INET_PRE_ROUTING,
    .pf = PF_INET6,
    .priority = NF_IP_PRI_FIRST,
};

static struct nf_hook_ops nfho_ipv6_post = {
    .hook = nf_hook_func,
    .hooknum = NF_INET_POST_ROUTING,
    .pf = PF_INET6,
    .priority = NF_IP_PRI_FIRST,
};

#define ENABLE_NETWORK_AUDIT  "enable network audit"
#define DISABLE_NETWORK_AUDIT "disable network audit"

static void nl_recv_msg(struct sk_buff *skb) {
    struct nlmsghdr *nlh;
    struct netlink_rcv *nc;

    enable_filter = 1;
    nlh = (struct nlmsghdr *)skb->data;
    
    //Check if the skb->data is NULL
    if(skb->data == NULL) {
        printk(KERN_INFO "skb->data is NULL\n");
        enable_filter = 0;
        return;
    }

    // Check if the nlh pointer content is "disable net_monitor" string
    if (strncmp(NLMSG_DATA(nlh), ENABLE_NETWORK_AUDIT, strlen(ENABLE_NETWORK_AUDIT)) == 0) {
        enable_filter = 1; // Set enable_filter to 1
    } else if(strncmp(NLMSG_DATA(nlh), DISABLE_NETWORK_AUDIT, strlen(DISABLE_NETWORK_AUDIT)) == 0) {
        enable_filter = 0; // Set enable_filter to 0
    } else {
        struct netlink_rcv *tmp = kzalloc(sizeof(struct netlink_rcv), GFP_KERNEL);
        int i;
        int white_count = 0;
        int black_count = 1;
        nc = (struct netlink_rcv *)NLMSG_DATA(nlh);

        // just truncate it when inputing too many rules
        size_t length = nc->length;
        if (length > MAX_RULES_LEN)
            length = MAX_RULES_LEN;

        //将 WHITE_LIST_RULES放在最前面,其他的放在后面
        for(i = 0; i < length; i++) {
            if(nc->rules[i].flags & NET_WHITELIST) {
                memcpy(&tmp->rules[white_count], &nc->rules[i], sizeof(nc->rules[i]));
                white_count++;
            } else {
                memcpy(&tmp->rules[length - black_count], &nc->rules[i], sizeof(nc->rules[i]));
                black_count++;
            }
        }
        tmp->length = length;

        print_netc(tmp);
        write_lock_bh(&my_rwlock);
        memcpy(&tuple_filter->netc, tmp, sizeof(struct netlink_rcv));
        write_unlock_bh(&my_rwlock);

        kfree(tmp);
    }
}

struct netlink_kernel_cfg cfg = {
    .input = nl_recv_msg,
};

// 初始化模块
static int __init nfhook_init(void) {
    int ret;

    enable_filter = 0; 
    tuple_filter = kzalloc(sizeof(struct filter), GFP_KERNEL); 
     
    // 注册Netfilter ipv4 钩子
    ret = nf_register_net_hook(&init_net,&nfho_ipv4_pre);
    if (ret < 0) {
        printk(KERN_ERR "Failed to register ipv4 pre nfhook\n");
        return ret;
    }

    ret = nf_register_net_hook(&init_net,&nfho_ipv4_post);
    if (ret < 0) {
        printk(KERN_ERR "Failed to register ipv4 post nfhook\n");
        return ret;
    }

    // 注册Netfilter ipv6 钩子
    ret = nf_register_net_hook(&init_net,&nfho_ipv6_pre);
    if (ret < 0) {
        printk(KERN_ERR "Failed to register ipv6 pre nfhook\n");
        return ret;
    }

    // 注册Netfilter ipv6 钩子
    ret = nf_register_net_hook(&init_net,&nfho_ipv6_post);
    if (ret < 0) {
        printk(KERN_ERR "Failed to register ipv6 pre nfhook\n");
        return ret;
    }

    nl_sk = netlink_kernel_create(&init_net, NETLINK_SOCKET, &cfg);
    if (!nl_sk) {
        printk(KERN_ERR "Failed to create netlink socket\n");
        return -1;
    }

    printk(KERN_INFO "nfhook module loaded\n");
    return 0;
}

// 卸载模块
static void __exit nfhook_exit(void) {
    
    nf_unregister_net_hook(&init_net,&nfho_ipv4_pre);
    nf_unregister_net_hook(&init_net,&nfho_ipv4_post);   
    nf_unregister_net_hook(&init_net,&nfho_ipv6_pre);
    nf_unregister_net_hook(&init_net,&nfho_ipv6_post);   

    if (nl_sk)
        netlink_kernel_release(nl_sk);

    printk(KERN_INFO "nfhook module unloaded\n");
}
module_init(nfhook_init);
module_exit(nfhook_exit);

MODULE_AUTHOR("yushengjin@uniontech.com");
MODULE_DESCRIPTION("net monitor module");
MODULE_LICENSE("GPL");
